body {
	font-family: "Segoe UI",Roboto,"Helvetica Neue",Arial,sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol";
	font-size:calc(14px + .75vw);
	padding:5vh 5vw;
	line-height:1.3;
}
h1 {
	font-size: calc(30px + 1.5vw);
	padding:1em 0 .68em;
	margin:0;
	color:#a3bac4;
}
#logo {
	width:max(38vw, 300px);
	margin-left:-.25vw;
}
